@@ -1,8 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
- * Copyright (C) 2003-2004 Imendio HB
- * Copyright (C) 2002-2003 Richard Hult <richard@imendio.com>
- * Copyright (C) 2002-2003 Mikael Hallendal <micke@imendio.com>
+ * Copyright (C) 2002-2004 Imendio HB
  * Copyright (C) 2003-2004 Geert-Jan Van den Bogaerde <gvdbogaerde@pandora.be>
  *
  * This program is free software; you can redistribute it and/or
@@ -38,10 +36,12 @@
 #include "gossip-contact-info.h"
 #include "gossip-stock.h"
 #include "gossip-log.h"
+#include "gossip-marshal.h"
 #include "gossip-roster.h"
 #include "gossip-chat.h"
 
 #define d(x)
+
 #define IS_ENTER(v) (v == GDK_Return || v == GDK_ISO_Enter || v == GDK_KP_Enter)
 #define COMPOSING_STOP_TIMEOUT 5
 
@@ -58,7 +58,7 @@ struct _GossipChatPriv {
 
 	GossipChatView   *view;
 
-        GossipRosterItem *item;
+        GossipContact    *contact;
 	gchar            *name;
 	
 	gchar            *locked_resource;
@@ -71,11 +71,19 @@ struct _GossipChatPriv {
 	gchar            *composing_resource; 
 	
 	gboolean          is_online;
+	gboolean          groupchat_priv;
+
+	/* Used to automatically shrink a window that has temporary grown
+	 * due to long input */
+	gint              padding_height;
+	gint              default_window_height;
+	gint              last_input_height;
+	gboolean          is_first_char;
 };
 
-static void            gossip_chat_class_init            (GossipChatClass  *klass);
-static void            gossip_chat_init                  (GossipChat       *chat);
-static void            gossip_chat_finalize              (GObject          *object);
+static void            chat_class_init                   (GossipChatClass  *klass);
+static void            chat_init                         (GossipChat       *chat);
+static void            chat_finalize                     (GObject          *object);
 static void            chats_init                        (void);
 static void            chat_create_gui                   (GossipChat       *chat);
 static void            chat_send                         (GossipChat       *chat,
@@ -91,17 +99,17 @@ static LmHandlerResult chat_message_handler              (LmMessageHandler *hand
                                                           LmConnection     *connection,
                                                           LmMessage        *m,
                                                           gpointer          user_data);
-static void            chat_item_presence_updated        (gpointer          not_used,
-							  GossipRosterItem *item,
+static void            chat_contact_presence_updated     (gpointer          not_used,
+							  GossipContact    *contact,
 							  GossipChat       *chat);
-static void            chat_item_updated                 (gpointer          not_used,
-							  GossipRosterItem *item,
+static void            chat_contact_updated              (gpointer          not_used,
+							  GossipContact    *contact,
 							  GossipChat       *chat);
-static void            chat_item_removed                 (gpointer          not_used,
-							  GossipRosterItem *item,
+static void            chat_contact_removed              (gpointer          not_used,
+							  GossipContact    *contact,
 							  GossipChat       *chat);
-static void            chat_item_added			 (gpointer	    not_used,
-							  GossipRosterItem *item,
+static void            chat_contact_added		 (gpointer	    not_used,
+							  GossipContact    *contact,
 							  GossipChat       *chat);
 static void            chat_connected_cb		 (GossipApp	   *app,
 							  GossipChat	   *chat);
@@ -119,6 +127,9 @@ static void            chat_input_text_buffer_changed_cb (GtkTextBuffer    *buff
 static gboolean	       chat_text_view_focus_in_event_cb  (GtkWidget        *widget,
 							  GdkEvent	   *event,
 							  GossipChat       *chat);
+static void            chat_text_view_size_allocate_cb   (GtkWidget        *widget,
+							  GtkAllocation    *allocation,
+							  GossipChat       *chat);
 static gboolean	       chat_delete_event_cb		 (GtkWidget	   *widget,
 							  GdkEvent	   *event,
 							  GossipChat       *chat);
@@ -144,12 +155,12 @@ gossip_chat_get_type (void)
                         sizeof (GossipChatClass),
                         NULL,
                         NULL,
-                        (GClassInitFunc) gossip_chat_class_init,
+                        (GClassInitFunc) chat_class_init,
                         NULL,
                         NULL,
                         sizeof (GossipChat),
                         0,
-                        (GInstanceInitFunc) gossip_chat_init
+                        (GInstanceInitFunc) chat_init
                 };
 
                 type_id = g_type_register_static (G_TYPE_OBJECT,
@@ -162,12 +173,13 @@ gossip_chat_get_type (void)
 }
 
 static void
-gossip_chat_class_init (GossipChatClass *klass)
+chat_class_init (GossipChatClass *klass)
 {
         GObjectClass *object_class = G_OBJECT_CLASS (klass);
+
         parent_class = g_type_class_peek_parent (klass);
 
-        object_class->finalize = gossip_chat_finalize;
+        object_class->finalize = chat_finalize;
 
         chat_signals[COMPOSING] =
                 g_signal_new ("composing",
@@ -175,7 +187,7 @@ gossip_chat_class_init (GossipChatClass *klass)
                               G_SIGNAL_RUN_LAST,
                               G_STRUCT_OFFSET (GossipChatClass, composing),
                               NULL, NULL,
-                              g_cclosure_marshal_VOID__BOOLEAN,
+			      gossip_marshal_VOID__BOOLEAN,
                               G_TYPE_NONE,
                               1,
                               G_TYPE_BOOLEAN);
@@ -186,40 +198,40 @@ gossip_chat_class_init (GossipChatClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GossipChatClass, new_message),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__VOID,
+			      gossip_marshal_VOID__VOID,
 			      G_TYPE_NONE,
 			      0);
 
-	chat_signals[NAME_CHANGED]=
+	chat_signals[NAME_CHANGED] =
 		g_signal_new ("name-changed",
 			      G_OBJECT_CLASS_TYPE (object_class),
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (GossipChatClass, name_changed),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__POINTER,
+			      gossip_marshal_VOID__POINTER,
 			      G_TYPE_NONE,
 			      1, G_TYPE_POINTER);
 }
 
 static void
-gossip_chat_init (GossipChat *chat)
+chat_init (GossipChat *chat)
 {
         GossipChatPriv   *priv;
 	LmConnection     *connection;
-	GossipRoster     *roster;
 	LmMessageHandler *handler;
 	
 	priv = g_new0 (GossipChatPriv, 1);
 	priv->window = NULL;
 	priv->request_composing_events = TRUE;
 	priv->is_online = FALSE;
-	
+	priv->is_first_char = TRUE;
+	priv->default_window_height = -1;
+
 	chat->priv = priv;
 	
         chat_create_gui (chat);
 
 	connection = gossip_app_get_connection ();
-	roster = gossip_app_get_roster ();
 
 	handler = lm_message_handler_new (chat_message_handler, chat, NULL);
 	chat->priv->message_handler = handler;
@@ -228,39 +240,19 @@ gossip_chat_init (GossipChat *chat)
 						LM_MESSAGE_TYPE_MESSAGE,
 						LM_HANDLER_PRIORITY_NORMAL);
 
-	g_signal_connect (roster,
-			  "item_presence_updated",
-			  G_CALLBACK (chat_item_presence_updated),
-			  chat);
-
-	g_signal_connect (roster,
-			  "item_updated",
-			  G_CALLBACK (chat_item_updated),
-			  chat);
-
-	g_signal_connect (roster,
-			  "item_removed",
-			  G_CALLBACK (chat_item_removed),
-			  chat);
-
-	g_signal_connect (roster,
-			  "item_added",
-			  G_CALLBACK (chat_item_added),
-			  chat);
-
-	g_signal_connect (gossip_app_get (), 
-			  "connected",
-			  G_CALLBACK (chat_connected_cb),
-			  chat);
-
-	g_signal_connect (gossip_app_get (), 
-			  "disconnected",
-			  G_CALLBACK (chat_disconnected_cb),
-			  chat);
+	g_signal_connect_object (gossip_app_get (), 
+				 "connected",
+				 G_CALLBACK (chat_connected_cb),
+				 chat, 0);
+	
+	g_signal_connect_object (gossip_app_get (), 
+				 "disconnected",
+				 G_CALLBACK (chat_disconnected_cb),
+				 chat, 0);
 }
 
 static void
-gossip_chat_finalize (GObject *object)
+chat_finalize (GObject *object)
 {
         GossipChat       *chat = GOSSIP_CHAT (object);
 	GossipChatPriv   *priv;
@@ -277,7 +269,7 @@ gossip_chat_finalize (GObject *object)
                 lm_message_handler_unref (handler);
         }
 
-	gossip_roster_item_unref (priv->item);
+	gossip_contact_unref (priv->contact);
 
 	chat_composing_remove_timeout (chat);
 	
@@ -300,9 +292,9 @@ chats_init (void)
 
         inited = TRUE;
 
-        chats = g_hash_table_new_full (gossip_jid_hash,
-				       gossip_jid_equal,
-                                       (GDestroyNotify) gossip_jid_unref,
+        chats = g_hash_table_new_full (gossip_contact_hash,
+				       gossip_contact_equal,
+                                       (GDestroyNotify) gossip_contact_unref,
                                        (GDestroyNotify) g_object_unref);
 }
 
@@ -348,6 +340,10 @@ chat_create_gui (GossipChat *chat)
 			  "focus_in_event",
 			  G_CALLBACK (chat_text_view_focus_in_event_cb),
 			  chat);
+	g_signal_connect (priv->input_text_view,
+			  "size_allocate",
+			  G_CALLBACK (chat_text_view_size_allocate_cb),
+			  chat);
 	g_signal_connect (priv->widget,
 			  "delete_event",
 			  G_CALLBACK (chat_delete_event_cb),
@@ -362,10 +358,13 @@ static void
 chat_update_locked_resource (GossipChat *chat)
 {
 	GossipChatPriv *priv = chat->priv;
-	GossipJID      *jid;
 	const gchar    *roster_resource;
 
-	if (gossip_roster_item_is_offline (priv->item)) {
+	if (priv->groupchat_priv) {
+		return;
+	}
+	
+	if (!gossip_contact_is_online (priv->contact)) {
 		g_free (priv->roster_resource);
 		priv->roster_resource = NULL;
 
@@ -375,9 +374,8 @@ chat_update_locked_resource (GossipChat *chat)
 		return;
 	}
 
-	jid = gossip_roster_item_get_jid (priv->item);
-
-	roster_resource = gossip_jid_get_resource (jid);
+	roster_resource = gossip_roster_get_active_resource (gossip_app_get_roster (),
+							     priv->contact);
 
 	/* It seems like some agents don't set a resource sometimes (ICQ for
 	 * example). I don't know if it's a bug or not, but we need to handle
@@ -427,8 +425,10 @@ chat_get_jid_with_resource (GossipChat *chat, const gchar *resource)
 	GossipChatPriv *priv = chat->priv;
 	GossipJID      *jid;
 
-	jid = gossip_roster_item_get_jid (priv->item);
+	jid = gossip_contact_get_jid (priv->contact);
 
+	d(g_print ("Getting jid with resource: %s\n", resource));
+	
 	if (resource) {
 		return g_strconcat (gossip_jid_get_without_resource (jid),
 				    "/", resource, NULL);
@@ -445,10 +445,11 @@ chat_send (GossipChat  *chat,
         LmMessage      *m;
         gchar          *nick;
 	gchar          *jid_string;
+	gchar          *username;
 
 	priv = chat->priv;
 
-	gossip_app_status_force_nonaway ();
+	gossip_app_force_non_away ();
 
         if (msg == NULL || msg[0] == '\0') {
                 return;
@@ -463,12 +464,15 @@ chat_send (GossipChat  *chat,
 
         nick = gossip_jid_get_part_name (gossip_app_get_jid ());
 
+	username = gossip_jid_get_part_name (gossip_app_get_jid ());
+
         gossip_chat_view_append_chat_message (priv->view,
                                               NULL,
-					      gossip_app_get_username (),
+					      username,
                                               nick,
                                               msg);
 
+	g_free (username);
         g_free (nick);
 
 	chat_update_locked_resource (chat);
@@ -483,7 +487,7 @@ chat_send (GossipChat  *chat,
 	
 	lm_message_node_add_child (m->node, "body", msg);
         
-        if (!gossip_roster_item_is_offline (priv->item) &&
+        if (gossip_contact_is_online (priv->contact) &&
 	    priv->request_composing_events) {
                 chat_request_composing (m);
         }
@@ -554,7 +558,7 @@ chat_composing_send_start_event (GossipChat *chat)
                 return;
         }
 
-	if (gossip_roster_item_is_offline (priv->item)) {
+	if (!gossip_contact_is_online (priv->contact)) {
 		return;
 	}
 	
@@ -678,15 +682,31 @@ chat_message_handler (LmMessageHandler *handler,
 
         from = lm_message_node_get_attribute (m->node, "from");
 
-        from_jid = gossip_jid_new (from);
-	jid = gossip_roster_item_get_jid (priv->item);
+	if (!from) {
+		g_print ("Received message without from attribute");
+		return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+	}
 
-        d(g_print ("Incoming message:: '%s' ?= '%s'\n",
-                   gossip_jid_get_without_resource (from_jid),
-                   gossip_jid_get_without_resource (jid)));
+        from_jid = gossip_jid_new (from);
+	jid = gossip_contact_get_jid (priv->contact);
+
+	if (lm_message_get_sub_type (m) == LM_MESSAGE_SUB_TYPE_GROUPCHAT) {
+		d(g_print ("GROUP CHAT!\n"));
+		gossip_jid_unref (from_jid);
+		
+                return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
+	}
 	
-        if (!gossip_jid_equals_without_resource (from_jid, jid)) {
-                gossip_jid_unref (from_jid);
+        d(g_print ("Incoming message:: '%s' ?= '%s'\n",
+                   gossip_jid_get_full (from_jid),
+                   gossip_jid_get_full (jid)));
+
+        if ((!priv->groupchat_priv && !gossip_jid_equals_without_resource (from_jid, jid)) ||
+
+	    (priv->groupchat_priv && !gossip_jid_equals (from_jid, jid))) {
+		d(g_print ("GROUP CHAT2!\n"));
+		gossip_jid_unref (from_jid);
+		
                 return LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
         }
 
@@ -716,7 +736,8 @@ chat_message_handler (LmMessageHandler *handler,
 
 	from_resource = gossip_jid_get_resource (from_jid);
 	
-	if (from_resource &&
+	if (!priv->groupchat_priv &&
+	    from_resource &&
 	    (!priv->locked_resource ||
 	     g_ascii_strcasecmp (from_resource, priv->locked_resource) != 0)) {
 
@@ -748,7 +769,7 @@ chat_message_handler (LmMessageHandler *handler,
         gossip_chat_view_append_chat_message (chat->priv->view,
                                               timestamp,
 					      NULL,
-                                              gossip_roster_item_get_name (priv->item),
+                                              gossip_contact_get_name (priv->contact),
                                               body);
 
 	g_signal_emit (chat, chat_signals[NEW_MESSAGE], 0);
@@ -765,23 +786,23 @@ chat_message_handler (LmMessageHandler *handler,
 }
 
 static void
-chat_item_presence_updated (gpointer          not_used,
-			    GossipRosterItem *item,
-			    GossipChat       *chat)
+chat_contact_presence_updated (gpointer       not_used,
+			       GossipContact *contact,
+			       GossipChat    *chat)
 
 {
 	GossipChatPriv *priv;
 
 	g_return_if_fail (GOSSIP_IS_CHAT (chat));
-	g_return_if_fail (item != NULL);
+	g_return_if_fail (contact != NULL);
 	
 	priv = chat->priv;
 
-	if (item != priv->item) {
+	if (!gossip_contact_equal (contact, priv->contact)) {
 		return;
 	}
 
-	if (gossip_roster_item_is_offline (item)) {
+	if (!gossip_contact_is_online (contact)) {
 		chat_composing_remove_timeout (chat);
 		priv->send_composing_events = FALSE;
 
@@ -792,8 +813,8 @@ chat_item_presence_updated (gpointer          not_used,
 			gchar *msg;
 			
 			msg = g_strdup_printf (_("%s went offline"),
-					       gossip_roster_item_get_name (priv->item));
-			gossip_chat_view_append_event_msg (priv->view,
+					       gossip_contact_get_name (priv->contact));
+			gossip_chat_view_append_event_msg (priv->view, 
 							   msg, TRUE);
 			g_free (msg);
 		}
@@ -806,7 +827,7 @@ chat_item_presence_updated (gpointer          not_used,
 			gchar *msg;
 		
 			msg = g_strdup_printf (_("%s comes online"),
-					       gossip_roster_item_get_name (priv->item));
+					       gossip_contact_get_name (priv->contact));
 			gossip_chat_view_append_event_msg (priv->view,
 							   msg, TRUE);
 			g_free (msg);
@@ -816,73 +837,67 @@ chat_item_presence_updated (gpointer          not_used,
 }
 
 static void
-chat_item_updated (gpointer          not_used,
-		   GossipRosterItem *item,
-		   GossipChat       *chat)
+chat_contact_updated (gpointer       not_used, 
+		      GossipContact *contact, 
+		      GossipChat    *chat)
 {
 	GossipChatPriv *priv;
 
 	g_return_if_fail (GOSSIP_IS_CHAT (chat));
-	g_return_if_fail (item != NULL);
+	g_return_if_fail (contact != NULL);
 	
 	priv = chat->priv;
 
-	if (item != priv->item) {
+	if (!gossip_contact_equal (contact, priv->contact)) {
 		return;
 	}
 
-	if (strcmp (priv->name, gossip_roster_item_get_name (item)) != 0) {
+	if (strcmp (priv->name, gossip_contact_get_name (contact)) != 0) {
 		g_free (priv->name);
-		priv->name = g_strdup (gossip_roster_item_get_name (item));
+		priv->name = g_strdup (gossip_contact_get_name (contact));
 		g_signal_emit (chat, chat_signals[NAME_CHANGED], 0, priv->name);
 	}
 }
 
 static void 
-chat_item_removed (gpointer          not_used,
-		   GossipRosterItem *item,
-		   GossipChat       *chat)
+chat_contact_removed (gpointer       not_used,
+		      GossipContact *contact,
+		      GossipChat    *chat)
 {
 	GossipChatPriv *priv;
 
 	g_return_if_fail (GOSSIP_IS_CHAT (chat));
-	g_return_if_fail (item != NULL);
+	g_return_if_fail (contact != NULL);
 	
 	priv = chat->priv;
 
-	if (item != priv->item) {
+	if (!gossip_contact_equal (contact, priv->contact)) {
 		return;
 	}
 }
 
 static void
-chat_item_added (gpointer          not_user,
-		 GossipRosterItem *item,
-		 GossipChat       *chat)
+chat_contact_added (gpointer       not_user,
+		    GossipContact *contact,
+		    GossipChat    *chat)
 {
 	GossipChatPriv *priv;
-	GossipJID      *new_jid;
-	GossipJID      *old_jid;
 
 	g_return_if_fail (GOSSIP_IS_CHAT (chat));
-	g_return_if_fail (item != NULL);
+	g_return_if_fail (contact != NULL);
 
 	priv = chat->priv;
 	
-	new_jid = gossip_roster_item_get_jid (item);
-	old_jid = gossip_roster_item_get_jid (priv->item);
-
-	if (!gossip_jid_equals_without_resource (old_jid, new_jid))
+	if (!gossip_contact_equal (contact, priv->contact)) {
 		return;
-
-	gossip_roster_item_unref (priv->item);
-	priv->item = item;
-	gossip_roster_item_ref (item);
+	}
+	
+	gossip_contact_unref (priv->contact);
+	priv->contact = gossip_contact_ref (contact);
 }
 
 static void
-chat_connected_cb (GossipApp  *app,
-		   GossipChat *chat)
+chat_connected_cb (GossipApp *app, GossipChat *chat)
 {
 	GossipChatPriv *priv;
 
@@ -896,8 +911,7 @@ chat_connected_cb (GossipApp  *app,
 }
 
 static void
-chat_disconnected_cb (GossipApp  *app,
-		      GossipChat *chat)
+chat_disconnected_cb (GossipApp *app, GossipChat *chat)
 {
 	GossipChatPriv *priv;
 
@@ -1003,18 +1017,21 @@ chat_error_dialog (GossipChat  *chat,
 static void
 chat_input_text_view_send (GossipChat *chat)
 {
-	GtkTextBuffer *buffer;
-	GtkTextIter    start, end;
-	gchar	      *msg;
-	gboolean       send_normal_value;
+	GossipChatPriv *priv;
+	GtkTextBuffer  *buffer;
+	GtkTextIter     start, end;
+	gchar	       *msg;
+	gboolean        send_normal_value;
 
-	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (chat->priv->input_text_view));
+	priv = chat->priv;
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->input_text_view));
 
 	gtk_text_buffer_get_bounds (buffer, &start, &end);
 	msg = gtk_text_buffer_get_text (buffer, &start, &end, FALSE);
 
 	/* Clear the input field */
-	send_normal_value = chat->priv->send_composing_events;
+	send_normal_value = priv->send_composing_events;
 	chat->priv->send_composing_events = FALSE;
 	gtk_text_buffer_set_text (buffer, "", -1);
 	chat->priv->send_composing_events = send_normal_value;
@@ -1022,6 +1039,20 @@ chat_input_text_view_send (GossipChat *chat)
 	chat_send (chat, msg);
 
 	g_free (msg);
+
+	priv->is_first_char = TRUE;
+}
+
+void
+gossip_chat_scroll_down (GossipChat *chat)
+{
+	GossipChatPriv *priv;
+	
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+
+	gossip_chat_view_scroll_down (priv->view);
 }
 
 static gboolean
@@ -1029,16 +1060,33 @@ chat_input_key_press_event_cb (GtkWidget   *widget,
                                GdkEventKey *event,
                                GossipChat  *chat)
 {
-	GossipChatPriv *priv = chat->priv;
+	GossipChatPriv *priv;
 	GtkAdjustment  *adj;
 	gdouble         val;
-		
-	/* Catch enter but not ctrl-enter */
-	if (IS_ENTER (event->keyval) && !(event->state & GDK_CONTROL_MASK)) {
-		chat_input_text_view_send (chat);
+	
+	priv = chat->priv;
 
+	/* Catch enter but not ctrl/shift-enter */
+	if (IS_ENTER (event->keyval) && !(event->state & GDK_SHIFT_MASK)) {
+
+		/* This is to make sure that kinput2 gets the enter. And if 
+		 * it's handled there we shouldn't send on it. This is because
+		 * kinput2 uses Enter to commit letters. See:
+		 * http://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=104299
+		 */
+		if (gtk_im_context_filter_keypress (GTK_TEXT_VIEW (priv->input_text_view)->im_context, event)) {
+			GTK_TEXT_VIEW (priv->input_text_view)->need_im_reset = TRUE;
+			return TRUE;
+		}
+		
+		chat_input_text_view_send (chat);
 		return TRUE;
 	}
+	
+	if (IS_ENTER (event->keyval) && (event->state & GDK_SHIFT_MASK)) {
+		/* Newline for shift-enter. */
+		return FALSE;
+	}
 	else if (event->keyval == GDK_Page_Up) {
 		adj = gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (priv->text_view_sw));
 		gtk_adjustment_set_value (adj, adj->value - adj->page_size);
@@ -1057,16 +1105,41 @@ chat_input_key_press_event_cb (GtkWidget   *widget,
 }
 
 static void
-chat_input_text_buffer_changed_cb (GtkTextBuffer *buffer,
-                                   GossipChat    *chat)
+chat_input_text_buffer_changed_cb (GtkTextBuffer *buffer, GossipChat *chat)
 {
-        if (chat->priv->send_composing_events) {
+	GossipChatPriv *priv;
+
+	priv = chat->priv;
+
+        if (priv->send_composing_events) {
                 if (gtk_text_buffer_get_char_count (buffer) == 0) {
-                        chat_composing_stop (chat);
+			chat_composing_stop (chat);
                 } else {
                         chat_composing_start (chat);
                 }
         }
+
+	if (priv->is_first_char) {
+		GtkRequisition  req;
+		gint            window_height;
+		GtkWidget      *dialog;
+		GtkAllocation  *allocation;
+		
+		/* Save the window's size */
+		dialog = gossip_chat_window_get_dialog (priv->window);
+		gtk_window_get_size (GTK_WINDOW (dialog), 
+				     NULL, &window_height);
+
+		gtk_widget_size_request (priv->input_text_view, &req);
+
+		allocation = &GTK_WIDGET (priv->view)->allocation;
+		
+		priv->default_window_height = window_height;
+		priv->last_input_height = req.height;
+		priv->padding_height = window_height - req.height - allocation->height;
+
+		chat->priv->is_first_char = FALSE;
+	}
 }
 
 static gboolean
@@ -1075,21 +1148,86 @@ chat_text_view_focus_in_event_cb (GtkWidget  *widget,
 				  GossipChat *chat)
 {
 	GossipChatPriv *priv = chat->priv;
-/*
-	gint pos;
 
-	pos = gtk_editable_get_position (GTK_EDITABLE (chat->priv->input_entry));
-
-	gtk_widget_grab_focus (chat->priv->input_entry);
-	gtk_editable_select_region (GTK_EDITABLE (chat->priv->input_entry), 0, 0);
-
-	gtk_editable_set_position (GTK_EDITABLE (chat->priv->input_entry), pos);
-*/
 	gtk_widget_grab_focus (priv->input_text_view);
 	
 	return TRUE;
 }
 
+typedef struct {
+	GtkWidget *window;
+	gint       width;
+	gint       height;
+} ChangeSizeData;
+
+static gboolean
+chat_change_size_in_idle_cb (ChangeSizeData *data)
+{
+	gtk_window_resize (GTK_WINDOW (data->window), 
+			   data->width, data->height);
+
+	return FALSE;
+}
+
+static void
+chat_text_view_size_allocate_cb (GtkWidget     *widget,
+				 GtkAllocation *allocation,
+				 GossipChat    *chat)
+{
+	GossipChatPriv *priv;
+	gint            width;
+	GtkWidget      *dialog;
+	ChangeSizeData *data;
+	gint            window_height;
+	gint            new_height;
+	GtkAllocation  *view_allocation;
+	gint            current_height;
+	gint            diff;
+	
+	priv = chat->priv;
+
+	if (priv->default_window_height <= 0) {
+		return;
+	}
+
+	if (priv->last_input_height <= allocation->height) {
+		priv->last_input_height = allocation->height;
+		return;
+	} 
+
+	diff = priv->last_input_height - allocation->height;
+	priv->last_input_height = allocation->height;
+
+	view_allocation = &GTK_WIDGET (priv->view)->allocation;
+
+	dialog = gossip_chat_window_get_dialog (priv->window);
+	gtk_window_get_size (GTK_WINDOW (dialog), NULL, &current_height);
+
+	new_height = view_allocation->height + priv->padding_height + allocation->height - diff;
+
+	if (new_height <= priv->default_window_height) {
+		window_height = priv->default_window_height;
+	} else {
+		window_height = new_height;
+	}
+
+	if (current_height <= window_height) {
+		return;
+	}
+
+	/* Restore the window's size */
+	gtk_window_get_size (GTK_WINDOW (dialog), &width, NULL);
+
+	data = g_new0 (ChangeSizeData, 1);
+	data->window = dialog;
+	data->width = width;
+	data->height = window_height;
+
+	g_idle_add_full (G_PRIORITY_DEFAULT_IDLE, 
+			 (GSourceFunc) chat_change_size_in_idle_cb,
+			 data, g_free);
+}
+
 static gboolean
 chat_delete_event_cb (GtkWidget  *widget,
 		      GdkEvent   *event,
@@ -1099,42 +1237,67 @@ chat_delete_event_cb (GtkWidget  *widget,
 }
 
 GossipChat *
-gossip_chat_get_for_item (GossipRosterItem *item, gboolean create)
+gossip_chat_get_for_contact (GossipContact *contact, gboolean create)
 {
 	GossipChat     *chat;
 	GossipChatPriv *priv;
-	GossipJID      *jid;
+	GossipRoster   *roster;
 
 	chats_init ();
 
-	jid = gossip_roster_item_get_jid (item);
-	chat = g_hash_table_lookup (chats, jid);
+	chat = g_hash_table_lookup (chats, contact);
 
 	if (chat) {
 		return chat;
 	}
 
-	if (!create)
+	if (!create) {
 		return NULL;
+	}
 	
 	chat = g_object_new (GOSSIP_TYPE_CHAT, NULL);
-	g_hash_table_insert (chats, gossip_jid_ref (jid), chat);
+	g_hash_table_insert (chats, gossip_contact_ref (contact), chat);
 	
 	priv = chat->priv;
-	priv->item = gossip_roster_item_ref (item);
-	priv->name = g_strdup (gossip_roster_item_get_name (item));
+	priv->contact = gossip_contact_ref (contact);
+	priv->name = g_strdup (gossip_contact_get_name (contact));
 
-	if (gossip_roster_item_is_offline (priv->item)) {
-		priv->is_online = FALSE;
-	} else {
+	priv->groupchat_priv = FALSE;
+
+	roster = gossip_app_get_roster ();
+
+	g_signal_connect_object (roster,
+				 "contact_presence_updated",
+				 G_CALLBACK (chat_contact_presence_updated),
+				 chat, 0);
+
+	g_signal_connect_object (roster,
+				 "contact_updated",
+				 G_CALLBACK (chat_contact_updated),
+				 chat, 0);
+
+	g_signal_connect_object (roster,
+				 "contact_removed",
+				 G_CALLBACK (chat_contact_removed),
+				 chat, 0);
+	
+	g_signal_connect_object (roster,
+				 "contact_added",
+				 G_CALLBACK (chat_contact_added),
+				 chat, 0);
+
+	if (gossip_contact_is_online (priv->contact)) {
 		priv->is_online = TRUE;
+	} else {
+		priv->is_online = FALSE;
 	}
 
 	return chat;
 }
 
 GossipChat *
-gossip_chat_get_for_group_chat (GossipRosterItem *item)
+gossip_chat_get_for_group_chat (GossipContact   *contact, 
+				GossipGroupChat *g_chat)
 {
 	GossipChat     *chat;
 	GossipChatPriv *priv;
@@ -1142,22 +1305,49 @@ gossip_chat_get_for_group_chat (GossipRosterItem *item)
 
 	chats_init ();
 
-	jid = gossip_roster_item_get_jid (item);
-	chat = g_hash_table_lookup (chats, jid);
+	chat = g_hash_table_lookup (chats, contact);
 	if (chat) {
 		return chat;
 	}
-
-	chat = g_object_new (GOSSIP_TYPE_CHAT, NULL);
-	priv = chat->priv;
-	priv->item = gossip_roster_item_ref (item);
 	
-	if (gossip_roster_item_is_offline (priv->item)) {
-		priv->is_online = FALSE;
-	} else {
+	jid = gossip_contact_get_jid (contact);
+	
+	chat = g_object_new (GOSSIP_TYPE_CHAT, NULL);
+	g_hash_table_insert (chats, gossip_contact_ref (contact), chat);
+
+	priv = chat->priv;
+	priv->contact = gossip_contact_ref (contact);
+	priv->name = g_strdup (gossip_contact_get_name (contact));
+	priv->groupchat_priv = TRUE;
+
+	priv->locked_resource = g_strdup (gossip_jid_get_resource (jid));
+	
+	if (gossip_contact_is_online (priv->contact)) {
 		priv->is_online = TRUE;
+	} else {
+		priv->is_online = FALSE;
 	}
 
+	g_signal_connect_object (g_chat,
+				 "contact_presence_updated",
+				 G_CALLBACK (chat_contact_presence_updated),
+				 chat, 0);
+
+	g_signal_connect_object (g_chat,
+				 "contact_updated",
+				 G_CALLBACK (chat_contact_updated),
+				 chat, 0);
+
+	g_signal_connect_object (g_chat,
+				 "contact_removed",
+				 G_CALLBACK (chat_contact_removed),
+				 chat, 0);
+	
+	g_signal_connect_object (g_chat,
+				 "contact_added",
+				 G_CALLBACK (chat_contact_added),
+				 chat, 0);
+	
 	return chat;
 }
 
@@ -1208,6 +1398,7 @@ gossip_chat_handle_message (LmMessage *m)
         GossipChat       *chat;
         LmHandlerResult   result;
 	GossipRosterItem *item;
+	GossipContact    *contact;
 
 	chats_init ();
 
@@ -1215,23 +1406,30 @@ gossip_chat_handle_message (LmMessage *m)
         jid = gossip_jid_new (from);
 
 	item = gossip_roster_get_item (gossip_app_get_roster (), jid);
-	if (!item) {
-		item = gossip_roster_item_new (jid);
+	if (item) {
+		contact = gossip_roster_get_contact_from_item (gossip_app_get_roster (),
+							       item);
+		gossip_contact_ref (contact);
+	} else {
+		contact = gossip_contact_new (GOSSIP_CONTACT_TYPE_TEMPORARY);
+		gossip_contact_set_jid (contact, jid);
 	}
-	chat = (GossipChat *) g_hash_table_lookup (chats, jid);
+
+	gossip_jid_unref (jid);
+
+	chat = (GossipChat *) g_hash_table_lookup (chats, contact);
 
 	if (chat) {
 		/* The existing message handler will catch it. */
 		result = LM_HANDLER_RESULT_ALLOW_MORE_HANDLERS;
 	} else {
-		chat = gossip_chat_get_for_item (item, TRUE);
-		/* chat = g_object_new (GOSSIP_TYPE_CHAT,
-				     "jid", jid, NULL); */
+		chat = gossip_chat_get_for_contact (contact, TRUE);
+		
 		gossip_chat_append_message (chat, m);
 		result = LM_HANDLER_RESULT_REMOVE_MESSAGE;
 	}
 
-        gossip_jid_unref (jid);
+	gossip_contact_unref (contact);
 
         return result;
 }
@@ -1262,8 +1460,8 @@ gossip_chat_get_widget (GossipChat *chat)
         return chat->priv->widget;
 }
 
-GossipRosterItem *
-gossip_chat_get_item (GossipChat *chat)
+GossipContact *
+gossip_chat_get_contact (GossipChat *chat)
 {
 	GossipChatPriv *priv;
 	
@@ -1271,7 +1469,7 @@ gossip_chat_get_item (GossipChat *chat)
 	
 	priv = chat->priv;
 
-	return priv->item;
+	return priv->contact;
 }
 
 void
@@ -1298,3 +1496,65 @@ gossip_chat_clear (GossipChat *chat)
 	gossip_chat_view_clear (priv->view);
 }
 
+void 
+gossip_chat_cut (GossipChat *chat)
+{
+	GossipChatPriv *priv;
+	GtkTextBuffer  *buffer;
+	
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+	
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->input_text_view));
+	if (gtk_text_buffer_get_selection_bounds (buffer, NULL, NULL)) {
+		GtkClipboard *clipboard;
+
+		clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+
+		gtk_text_buffer_cut_clipboard (buffer, clipboard, TRUE);
+	}
+}
+
+void 
+gossip_chat_copy (GossipChat *chat)
+{
+	GossipChatPriv *priv;
+	GtkTextBuffer  *buffer;
+	
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+	
+	if (gossip_chat_view_get_selection_bounds (priv->view, NULL, NULL)) {
+		gossip_chat_view_copy_clipboard (priv->view);
+		return;
+	}
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->input_text_view));
+	if (gtk_text_buffer_get_selection_bounds (buffer, NULL, NULL)) {
+		GtkClipboard *clipboard;
+
+		clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+
+		gtk_text_buffer_copy_clipboard (buffer, clipboard);
+	}
+}
+
+void
+gossip_chat_paste (GossipChat *chat)
+{
+	GossipChatPriv *priv;
+	GtkTextBuffer  *buffer;
+	GtkClipboard   *clipboard;
+
+	g_return_if_fail (GOSSIP_IS_CHAT (chat));
+
+	priv = chat->priv;
+
+	buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (priv->input_text_view));
+	clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+
+	gtk_text_buffer_paste_clipboard (buffer, clipboard, NULL, TRUE);
+}
+
