@@ -34,12 +34,16 @@
 #include "gossip-app.h"
 #include "gossip-utils.h"
 
+#define DINGUS "(((mailto|news|telnet|nttp|file|http|sftp|ftp|https|dav|callto)://)|(www|ftp)[-A-Za-z0-9]*\\.)[-A-Za-z0-9\\.]+(:[0-9]*)?(/[-A-Za-z0-9_\\$\\.\\+\\!\\*\\(\\),;:{}@&=\\?/~\\#\\%]*[^]'\\.}>\\) ,\\\"])?"
+
 #define AVAILABLE_MESSAGE "Available"
 #define AWAY_MESSAGE "Away"
 #define BUSY_MESSAGE "Busy"
 
 extern GConfClient *gconf_client;
 
+static regex_t  dingus;
+
 typedef enum {
     TIMESTAMP_STYLE_IRC,
     TIMESTAMP_STYLE_NORMAL
@@ -470,28 +474,6 @@ gossip_password_dialog_run (GossipAccount *account, GtkWindow *parent)
 	return password;
 }
 
-const gchar *
-gossip_get_icon_for_show_string (const gchar *str)
-{
-	if (!str || !str[0]) {
-		return GOSSIP_STOCK_AVAILABLE;
-	}
-	else if (strcmp (str, "dnd") == 0) {
-		return GOSSIP_STOCK_BUSY;
-	}
-	else if (strcmp (str, "away") == 0) {
-		return GOSSIP_STOCK_AWAY;
-	}
-	else if (strcmp (str, "xa") == 0) {
-		return GOSSIP_STOCK_EXT_AWAY;
-	}
-	/* We don't support chat, so treat it like available. */
-	else if (strcmp (str, "chat") == 0) {
-		return GOSSIP_STOCK_AVAILABLE;
-	}
-
-	return GOSSIP_STOCK_AVAILABLE;
-}
 
 const gchar *
 gossip_utils_show_to_string (GossipShow show)
@@ -763,15 +745,31 @@ gossip_utils_get_pixbuf_from_show (GossipShow show)
 }
 
 const gchar *
-gossip_utils_get_default_status (GossipShow show)
+gossip_utils_get_default_status_show (GossipShow show)
 {
 	switch (show) {
-	case GOSSIP_SHOW_AVAILABLE:
+        case GOSSIP_SHOW_AVAILABLE:
+                return _(AVAILABLE_MESSAGE);
+        case GOSSIP_SHOW_BUSY:
+                return _(BUSY_MESSAGE);
+        case GOSSIP_SHOW_AWAY:
+        case GOSSIP_SHOW_EXT_AWAY:
+                return _(AWAY_MESSAGE);
+        };
+ 
+        return _(AVAILABLE_MESSAGE);
+}
+
+const gchar *
+gossip_utils_get_default_status (GossipPresenceType type)
+{
+	switch (type) {
+	case GOSSIP_PRESENCE_TYPE_AVAILABLE:
 		return _(AVAILABLE_MESSAGE);
-	case GOSSIP_SHOW_BUSY:
+	case GOSSIP_PRESENCE_TYPE_BUSY:
 		return _(BUSY_MESSAGE);
-	case GOSSIP_SHOW_AWAY:
-	case GOSSIP_SHOW_EXT_AWAY:
+	case GOSSIP_PRESENCE_TYPE_AWAY:
+	case GOSSIP_PRESENCE_TYPE_EXT_AWAY:
 		return _(AWAY_MESSAGE);
 	};
 
@@ -879,3 +877,138 @@ gossip_utils_free_status_messages (GList *list)
 	g_list_foreach (list, (GFunc) free_entry, NULL);
 	g_list_free (list);
 }
+
+gint
+gossip_utils_url_regex_match (const gchar *msg,
+			      GArray      *start,
+			      GArray      *end)
+{
+	static gboolean inited = FALSE;
+	regmatch_t      matches[1];
+	gint            ret = 0;
+	gint            num_matches = 0;
+	gint            offset = 0;
+
+	if (!inited) {
+		memset (&dingus, 0, sizeof (regex_t));
+		regcomp (&dingus, DINGUS, REG_EXTENDED);
+		inited = TRUE;
+	}
+
+	while (!ret) {
+		ret = regexec (&dingus, msg + offset, 1, matches, 0);
+
+		if (ret == 0) {
+			gint s;
+			
+			num_matches++;
+
+			s = matches[0].rm_so + offset;
+			offset = matches[0].rm_eo + offset;
+			
+			g_array_append_val (start, s);
+			g_array_append_val (end, offset);
+		}
+	}
+		
+	return num_matches;
+}
+
+GossipPresenceType
+gossip_utils_get_presence_type_from_show_string (const gchar *str)
+{
+	if (!str || !str[0]) {
+		return GOSSIP_PRESENCE_TYPE_AVAILABLE;
+	}
+	else if (strcmp (str, "dnd") == 0) {
+		return GOSSIP_PRESENCE_TYPE_BUSY;
+	}
+	else if (strcmp (str, "away") == 0) {
+		return GOSSIP_PRESENCE_TYPE_AWAY;
+	}
+	else if (strcmp (str, "xa") == 0) {
+		return GOSSIP_PRESENCE_TYPE_EXT_AWAY;
+	}
+	/* We don't support chat, so treat it like available. */
+	else if (strcmp (str, "chat") == 0) {
+		return GOSSIP_PRESENCE_TYPE_AVAILABLE;
+	}
+
+	return GOSSIP_PRESENCE_TYPE_AVAILABLE;
+}
+
+gint
+gossip_utils_str_case_cmp (const gchar *s1, const gchar *s2)
+{
+	return gossip_utils_str_n_case_cmp (s1, s2, -1);
+}	
+
+gint
+gossip_utils_str_n_case_cmp (const gchar *s1, const gchar *s2, gsize n)
+{
+	gchar *u1, *u2;
+	gint   ret_val;
+	
+	u1 = g_utf8_casefold (s1, n);
+	u2 = g_utf8_casefold (s2, n);
+
+	ret_val = g_utf8_collate (u1, u2);
+	g_free (u1);
+	g_free (u2);
+
+	return ret_val;
+}
+
+#define CONF_HTTP_PROXY_PREFIX "/system/http_proxy"
+
+void
+gossip_utils_set_proxy (LmConnection *conn)
+{
+	gboolean  use_http_proxy;
+	
+	use_http_proxy = gconf_client_get_bool (gconf_client,
+						CONF_HTTP_PROXY_PREFIX "/use_http_proxy",
+						NULL);
+	if (use_http_proxy) {
+		LmProxy  *proxy;
+		gchar    *host;
+		gint      port;
+		gboolean  use_auth;
+			host = gconf_client_get_string (gconf_client,
+						CONF_HTTP_PROXY_PREFIX "/host",
+						NULL);
+
+		port = gconf_client_get_int (gconf_client,
+					     CONF_HTTP_PROXY_PREFIX "/port",
+					     NULL);
+
+		proxy = lm_proxy_new_with_server (LM_PROXY_TYPE_HTTP,
+						  host, (guint) port);
+		g_free (host);
+
+		lm_connection_set_proxy (conn, proxy);
+
+		use_auth = gconf_client_get_bool (gconf_client,
+						  CONF_HTTP_PROXY_PREFIX "/use_authentication",
+						  NULL);
+		if (use_auth) {
+			gchar *username;
+			gchar *password;
+
+			username = gconf_client_get_string (gconf_client,
+							    CONF_HTTP_PROXY_PREFIX "/authentication_user",
+							    NULL);
+			password = gconf_client_get_string (gconf_client,
+							    CONF_HTTP_PROXY_PREFIX "/authentication_password",
+							    NULL);
+			lm_proxy_set_username (proxy, username);
+			lm_proxy_set_password (proxy, password);
+
+			g_free (username);
+			g_free (password);
+		}
+		
+		lm_proxy_unref (proxy);
+	}
+
+}
